@@ -561,21 +561,17 @@ python_munge_path <- function(python) {
   if (is_conda_python(python)) {
     conda_info <- get_python_conda_info(python)
 
-    if (numeric_conda_version(conda_info$conda) >= "4.9") {
-
-      new_path <- conda_run(
-        "python",
-        c("-c", shQuote("import os; print(os.environ['PATH'])")),
-        conda = conda_info$conda,
-        envname = conda_info$root,
-        stdout = TRUE
-      )
-
-      old_path <- Sys.getenv("PATH")
-      Sys.setenv("PATH" = new_path)
-      return(old_path)
-    }
+    new_path <- conda_run2(
+      "python",
+      c("-c", shQuote("import os; print(os.environ['PATH'])")),
+      conda = conda_info$conda,
+      envname = conda_info$root,
+      intern = TRUE
+    )
 
+    old_path <- Sys.getenv("PATH")
+    Sys.setenv("PATH" = new_path)
+    return(old_path)
   }
 
   if (is_windows()) {
@@ -602,6 +598,37 @@ python_munge_path <- function(python) {
 
 }
 
+python_config_impl <- function(python) {
+
+  script <- system.file("config/config.py", package = "reticulate")
+  config <- system2(
+    command = python,
+    args    = shQuote(script),
+    stdout  = TRUE,
+    stderr  = FALSE
+  )
+
+  # check for error
+  status <- attr(config, "status")
+  if (!is.null(status)) {
+    errmsg <- attr(config, "errmsg")
+    stop("Error ", status, " occurred running ", python, ": ", errmsg)
+  }
+
+  # on macOS, if we see some variables referencing /Applications/Xcode.app
+  # but that doesn't actually exist, then redirect to default CLT
+  if (is_osx()) {
+    clt <- "/Library/Developer/CommandLineTools"
+    xcode <- "/Applications/Xcode.app/Contents/Developer"
+    if (file.exists(clt))
+      config <- gsub(xcode, clt, config, fixed = TRUE)
+  }
+
+  # return config
+  config
+
+}
+
 python_config <- function(python,
                           required_module = NULL,
                           python_versions = python,
@@ -641,20 +668,7 @@ python_config <- function(python,
   }
 
   # execute config script
-  config_script <- system.file("config/config.py", package = "reticulate")
-  config <- system2(
-    command = python,
-    args    = shQuote(config_script),
-    stdout  = TRUE,
-    stderr  = FALSE
-  )
-
-  # check for error
-  status <- attr(config, "status")
-  if (!is.null(status)) {
-    errmsg <- attr(config, "errmsg")
-    stop("Error ", status, " occurred running ", python, " ", errmsg)
-  }
+  config <- python_config_impl(python)
 
   # read output as dcf
   config_connection <- textConnection(config)
@@ -700,14 +714,6 @@ python_config <- function(python,
 
   } else {
 
-    # (note that the LIBRARY variable has the name of the static library)
-    python_libdir_config <- function(var) {
-      python_libdir <- config[[var]]
-      ext <- switch(Sys.info()[["sysname"]], Darwin = ".dylib", Windows = ".dll", ".so")
-      pattern <- paste0("^libpython", version, "d?m?", ext)
-      libpython <- list.files(python_libdir, pattern = pattern, full.names = TRUE)
-    }
-
     # default to NULL
     libpython <- NULL
 
@@ -781,14 +787,6 @@ python_config <- function(python,
   # check for required module
   required_module_path <- config$RequiredModulePath
 
-  # fix up libpython for macOS command line tools
-  if (is_osx() && length(libpython)) {
-    old <- "/Applications/Xcode.app/Contents/Developer"
-    new <- "/Library/Developer/CommandLineTools"
-    if (grepl(new, config$PythonPath, fixed = TRUE))
-      libpython <- gsub(old, new, libpython, fixed = TRUE)
-  }
-
   # return config info
   info <- list(
     python               = python,
